from __future__ import print_function
from argparse import ArgumentParser
from importlib import import_module
import os
from subprocess import CalledProcessError
import sys

from atticmatic.config import parse_configuration


DEFAULT_CONFIG_FILENAME_PATTERN = '/etc/{}/config'
DEFAULT_EXCLUDES_FILENAME_PATTERN = '/etc/{}/excludes'


def parse_arguments(command_name, *arguments):
    '''
    Given the name of the command with which this script was invoked and command-line arguments,
    parse the arguments and return them as an ArgumentParser instance. Use the command name to
    determine the default configuration and excludes paths.
    '''
    config_filename_default = DEFAULT_CONFIG_FILENAME_PATTERN.format(command_name)
    excludes_filename_default = DEFAULT_EXCLUDES_FILENAME_PATTERN.format(command_name)

    parser = ArgumentParser()
    parser.add_argument(
        '-c', '--config',
        dest='config_filename',
        default=config_filename_default,
        help='Configuration filename',
    )
    parser.add_argument(
        '--excludes',
        dest='excludes_filename',
        default=excludes_filename_default if os.path.exists(excludes_filename_default) else None,
        help='Excludes filename',
    )
    parser.add_argument(
        '-v', '--verbosity',
        type=int,
        help='Display verbose progress (1 for some, 2 for lots)',
    )

    return parser.parse_args(arguments)


def load_backend(command_name):
    '''
    Given the name of the command with which this script was invoked, return the corresponding
    backend module responsible for actually dealing with backups.
    '''
    backend_name = {
        'atticmatic': 'attic',
        'borgmatic': 'borg',
    }.get(command_name, 'attic')

    return import_module('atticmatic.backends.{}'.format(backend_name))


def main():
    try:
        command_name = os.path.basename(sys.argv[0])
        args = parse_arguments(command_name, *sys.argv[1:])
        backend = load_backend(command_name)
        config = parse_configuration(args.config_filename, backend.CONFIG_FORMAT)
        repository = config.location['repository']

        backend.create_archive(args.excludes_filename, args.verbosity, **config.location)
        backend.prune_archives(args.verbosity, repository, config.retention)
        backend.check_archives(args.verbosity, repository, config.consistency)
    except (ValueError, IOError, CalledProcessError) as error:
        print(error, file=sys.stderr)
        sys.exit(1)



from functools import partial

from atticmatic.backends import shared

# An atticmatic backend that supports Attic for actually handling backups.

COMMAND = 'attic'
CONFIG_FORMAT = shared.CONFIG_FORMAT

create_archive = partial(shared.create_archive, command=COMMAND)
prune_archives = partial(shared.prune_archives, command=COMMAND)
check_archives = partial(shared.check_archives, command=COMMAND)



from flexmock import flexmock

from atticmatic import command as module


def test_load_backend_with_atticmatic_command_should_return_attic_backend():
    backend = flexmock()
    (
        flexmock(module).should_receive('import_module').with_args('atticmatic.backends.attic')
        .and_return(backend).once()
    )

    assert module.load_backend('atticmatic') == backend


def test_load_backend_with_unknown_command_should_return_attic_backend():
    backend = flexmock()
    (
        flexmock(module).should_receive('import_module').with_args('atticmatic.backends.attic')
        .and_return(backend).once()
    )

    assert module.load_backend('unknownmatic') == backend


def test_load_backend_with_borgmatic_command_should_return_borg_backend():
    backend = flexmock()
    (
        flexmock(module).should_receive('import_module').with_args('atticmatic.backends.borg')
        .and_return(backend).once()
    )

    assert module.load_backend('borgmatic') == backend



from functools import partial

from atticmatic.config import Section_format, option
from atticmatic.backends import shared

# An atticmatic backend that supports Borg for actually handling backups.

COMMAND = 'borg'
CONFIG_FORMAT = (
    shared.CONFIG_FORMAT[0],  # location
    shared.CONFIG_FORMAT[1],  # retention
    Section_format(
        'consistency',
        (
            option('checks', required=False),
            option('check_last', required=False),
        ),
    )
)


create_archive = partial(shared.create_archive, command=COMMAND)
prune_archives = partial(shared.prune_archives, command=COMMAND)
check_archives = partial(shared.check_archives, command=COMMAND)



from datetime import datetime
import os
import platform
import subprocess

from atticmatic.config import Section_format, option
from atticmatic.verbosity import VERBOSITY_SOME, VERBOSITY_LOTS


# Common backend functionality shared by Attic and Borg. As the two backup
# commands diverge, these shared functions will likely need to be replaced
# with non-shared functions within atticmatic.backends.attic and
# atticmatic.backends.borg.


CONFIG_FORMAT = (
    Section_format(
        'location',
        (
            option('source_directories'),
            option('repository'),
        ),
    ),
    Section_format(
        'retention',
        (
            option('keep_within', required=False),
            option('keep_hourly', int, required=False),
            option('keep_daily', int, required=False),
            option('keep_weekly', int, required=False),
            option('keep_monthly', int, required=False),
            option('keep_yearly', int, required=False),
            option('prefix', required=False),
        ),
    ),
    Section_format(
        'consistency',
        (
            option('checks', required=False),
        ),
    )
)

def create_archive(excludes_filename, verbosity, source_directories, repository, command):
    '''
    Given an excludes filename (or None), a vebosity flag, a space-separated list of source
    directories, a local or remote repository path, and a command to run, create an attic archive.
    '''
    sources = tuple(source_directories.split(' '))
    exclude_flags = ('--exclude-from', excludes_filename) if excludes_filename else ()
    verbosity_flags = {
        VERBOSITY_SOME: ('--stats',),
        VERBOSITY_LOTS: ('--verbose', '--stats'),
    }.get(verbosity, ())

    full_command = (
        command, 'create',
        '{repo}::{hostname}-{timestamp}'.format(
            repo=repository,
            hostname=platform.node(),
            timestamp=datetime.now().isoformat(),
        ),
    ) + sources + exclude_flags + verbosity_flags

    subprocess.check_call(full_command)


def _make_prune_flags(retention_config):
    '''
    Given a retention config dict mapping from option name to value, tranform it into an iterable of
    command-line name-value flag pairs.

    For example, given a retention config of:

        {'keep_weekly': 4, 'keep_monthly': 6}

    This will be returned as an iterable of:

        (
            ('--keep-weekly', '4'),
            ('--keep-monthly', '6'),
        )
    '''
    return (
        ('--' + option_name.replace('_', '-'), str(retention_config[option_name]))
        for option_name, value in retention_config.items()
    )


def prune_archives(verbosity, repository, retention_config, command):
    '''
    Given a verbosity flag, a local or remote repository path, a retention config dict, and a
    command to run, prune attic archives according the the retention policy specified in that
    configuration.
    '''
    verbosity_flags = {
        VERBOSITY_SOME: ('--stats',),
        VERBOSITY_LOTS: ('--verbose', '--stats'),
    }.get(verbosity, ())

    full_command = (
        command, 'prune',
        repository,
    ) + tuple(
        element
        for pair in _make_prune_flags(retention_config)
        for element in pair
    ) + verbosity_flags

    subprocess.check_call(full_command)


DEFAULT_CHECKS = ('repository', 'archives')


def _parse_checks(consistency_config):
    '''
    Given a consistency config with a space-separated "checks" option, transform it to a tuple of
    named checks to run.

    For example, given a retention config of:

        {'checks': 'repository archives'}

    This will be returned as:

        ('repository', 'archives')

    If no "checks" option is present, return the DEFAULT_CHECKS. If the checks value is the string
    "disabled", return an empty tuple, meaning that no checks should be run.
    '''
    checks = consistency_config.get('checks', '').strip()
    if not checks:
        return DEFAULT_CHECKS

    return tuple(
        check for check in consistency_config['checks'].split(' ')
        if check.lower() not in ('disabled', '')
    )


def _make_check_flags(checks, check_last=None):
    '''
    Given a parsed sequence of checks, transform it into tuple of command-line flags.

    For example, given parsed checks of:

        ('repository',)

    This will be returned as:
    
        ('--repository-only',)

    Additionally, if a check_last value is given, a "--last" flag will be added. Note that only
    Borg supports this flag.
    '''
    last_flag = ('--last', check_last) if check_last else ()
    if checks == DEFAULT_CHECKS:
        return last_flag

    return tuple(
        '--{}-only'.format(check) for check in checks
    ) + last_flag


def check_archives(verbosity, repository, consistency_config, command):
    '''
    Given a verbosity flag, a local or remote repository path, a consistency config dict, and a
    command to run, check the contained attic archives for consistency.

    If there are no consistency checks to run, skip running them.
    '''
    checks = _parse_checks(consistency_config)
    check_last = consistency_config.get('check_last', None)
    if not checks:
        return

    verbosity_flags = {
        VERBOSITY_SOME: ('--verbose',),
        VERBOSITY_LOTS: ('--verbose',),
    }.get(verbosity, ())

    full_command = (
        command, 'check',
        repository,
    ) + _make_check_flags(checks, check_last) + verbosity_flags

    # The check command spews to stdout even without the verbose flag. Suppress it.
    stdout = None if verbosity_flags else open(os.devnull, 'w')

    subprocess.check_call(full_command, stdout=stdout)



from setuptools import setup, find_packages


VERSION = '0.1.5'


setup(
    name='atticmatic',
    version=VERSION,
    description='A wrapper script for Attic/Borg backup software that creates and prunes backups',
    author='Dan Helfman',
    author_email='witten@torsion.org',
    url='https://torsion.org/atticmatic',
    download_url='https://torsion.org/hg/atticmatic/archive/%s.tar.gz' % VERSION,
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Archiving :: Backup',
    ),
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'atticmatic = atticmatic.command:main',
            'borgmatic = atticmatic.command:main',
        ]
    },
    tests_require=(
        'flexmock',
        'nose',
    )
)
